#include "../Framework2/Direct3D/Direct3DApp.h"
#include "../Framework2/Util/Direct3DModel.h"

#define SIZE 64
#define DIST_SCALE (1.0f / 2000.0f)

class MainApp : public Direct3DApp {
public:
	void resetCamera();

	bool init();
	bool exit();

	bool load();
	bool unload();

	void drawLight(const vec3 &lightPos);
	void drawDepth(const vec3 &lightPos);
	void drawQuad();
	bool drawFrame();
protected:
	Direct3DModel *model;

	ShaderID lighting, dist, asciiShader;

	TextureID cubeTarget, depthTarget, rtColor, rtDepth;

	TextureID light, asciiTex;
	TextureID base;
};

void MainApp::resetCamera(){
	position = vec3(450, 50, 400);
	wx = 0;
	wy = 3 * PI / 4;
	wz = 0;
}

bool MainApp::init(){
	model = new Direct3DModel();
	model->loadFromFile("Map.hmdl");

	return true;
}

bool MainApp::exit(){
	delete model;
	return true;
}

bool MainApp::load(){
	if (caps.PixelShaderVersion < D3DPS_VERSION(2,0) || caps.VertexShaderVersion < D3DVS_VERSION(1,1)){
		String error("This demo requires pixel shader 2.0 and vertex shader 1.1.\n"
					 "Your card/driver only supports pixel shader ");
		error.sprintf("%d.%d", (caps.PixelShaderVersion  >> 8) & 255, caps.PixelShaderVersion  & 255);
		error += " and vertex shader ";
		error.sprintf("%d.%d", (caps.VertexShaderVersion >> 8) & 255, caps.VertexShaderVersion & 255);

		addToLog(error);
		return false;
	}

	setDefaultFont("../Textures/Fonts/Future.font", "../Textures/Fonts/Future.dds");

	if ((lighting = renderer->addShader("lighting.shd")) == SHADER_NONE) return false;
	if ((dist = renderer->addShader("dist.shd")) == SHADER_NONE) return false;
	if ((asciiShader = renderer->addShader("ascii.shd")) == SHADER_NONE) return false;

	if ((cubeTarget = renderer->addRenderCubemap(256, FORMAT_RGBA8)) == TEXTURE_NONE) return false;
	if ((depthTarget = renderer->addRenderDepth(256, 256, FORMAT_D16)) == TEXTURE_NONE) return false;
	
	if ((rtColor = renderer->addRenderTexture(SIZE, SIZE, FORMAT_RGBA8)) == TEXTURE_NONE) return false;
	if ((rtDepth = renderer->addRenderDepth(SIZE, SIZE, FORMAT_D16)) == TEXTURE_NONE) return false;

	if ((base = renderer->addTexture("../Textures/Wood.dds")) == TEXTURE_NONE) return false;
	if ((light = renderer->addTexture("../Textures/Particle.tga")) == TEXTURE_NONE) return false;

	model->uploadToVertexBuffer(dev);


	Image img;
	if (!img.loadImage("../Textures/font.tga"))	return false;
	img.toGrayScale();

	unsigned char *src = img.getImage();

	float avg[256];
	unsigned int i, j, a, b;
	for (j = 0; j < 16; j++){
		for (i = 0; i < 16; i++){
			avg[j * 16 + i] = 0;

			unsigned int base = j * 256 * 16 + i * 16;
			for (b = 0; b < 16; b++){
				for (a = 0; a < 16; a++){
					avg[j * 16 + i] += float(src[base + 256 * b + a]);
				}				
			}
			avg[j * 16 + i] *= 1.0f / (255.0f * 16 * 16);
		}
	}

	unsigned char *pixels = new unsigned char[16 * 16 * 256];
	unsigned char *dest = pixels;

	for (j = 0; j < 256; j++){
		float av = float(j) / 255.0f;

		unsigned int index = 0;
		float minDiff = 1e10f;
		for (i = 0; i < 256; i++){
			float d = fabsf(avg[i] - av);
			if (d < minDiff){
				minDiff = d;
				index = i;
			}			
		}

		unsigned int r = index / 16;
		unsigned int c = index % 16;
		unsigned int base = r * 256 * 16 + c * 16;
		for (b = 0; b < 16; b++){
			for (a = 0; a < 16; a++){
				*dest++ = src[base + 256 * b + a];
			}
		}
	}

	Image3D *img3D = new Image3D();
	img3D->loadFromMemory(pixels, 16, 16, 256, FORMAT_I8, true, false);

	asciiTex = renderer->addTexture3D(img3D, TEX_NOMIPMAPPING | TEX_CLAMP_R);

	return true;
}

bool MainApp::unload(){
	model->freeVertexBuffer();
	return true;
}

void MainApp::drawLight(const vec3 &lightPos){
	renderer->setTextures(light);
	renderer->setBlending(ONE, ONE);
	renderer->apply();

	vec3 dx(modelView.elem[0][0], modelView.elem[0][1], modelView.elem[0][2]);
	vec3 dy(modelView.elem[1][0], modelView.elem[1][1], modelView.elem[1][2]);

	struct Vertex {
		vec3 position;
		vec2 texCoord;
	};

	Vertex vertices[4];

	vertices[0].position = lightPos + 30 * (-dx + dy);
	vertices[1].position = lightPos + 30 * ( dx + dy);
	vertices[2].position = lightPos + 30 * ( dx - dy);
	vertices[3].position = lightPos + 30 * (-dx - dy);

	vertices[0].texCoord = vec2(0, 0);
	vertices[1].texCoord = vec2(1, 0);
	vertices[2].texCoord = vec2(1, 1);
	vertices[3].texCoord = vec2(0, 1);


	dev->SetFVF(D3DFVF_XYZ | D3DFVF_TEXCOORDSIZE2(0) | (1 << D3DFVF_TEXCOUNT_SHIFT));
	dev->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, vertices, sizeof(Vertex));
}

void MainApp::drawDepth(const vec3 &lightPos){
	renderer->setShader(dist);
	renderer->apply();

	renderer->changeShaderConstant3f("lightPos", lightPos);
	renderer->changeShaderConstant1f("distScale", DIST_SCALE);

	mat4 proj = projectionMatrixX(PI / 2, 1, 1, 3000);
	for (int i = 0; i < 6; i++){
		renderer->changeRenderTarget(cubeTarget, depthTarget, i);
		dev->Clear(0, NULL, D3DCLEAR_ZBUFFER, D3DCOLOR_RGBA(0, 0, 0, 0), 1.0f, 0);

		mat4 view = cubemapMatrix(i);
		view.translate(-lightPos);

		renderer->changeShaderConstant4x4f("view_proj_matrix", proj * view);
		model->draw(dev);
	}
}

void MainApp::drawQuad(){
	static float vertices[] = {
		-1,  1, 0, 0, 0,
		 1,  1, 0, 1, 0,
		 1, -1, 0, 1, 1,
		-1, -1, 0, 0, 1,
	};

	dev->SetFVF(D3DFVF_XYZ | D3DFVF_TEXCOORDSIZE2(0) | (1 << D3DFVF_TEXCOUNT_SHIFT));
	dev->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, vertices, 5 * sizeof(float));
}

bool MainApp::drawFrame(){
    //dev->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_RGBA(0, 0, 0, 0), 1.0f, 0);

	projection = projectionMatrixX(1.5f, float(height) / float(width), 1, 3000);
	dev->SetTransform(D3DTS_PROJECTION, (D3DMATRIX *) (const float *) transpose(projection));

	modelView = rotateZXY(-wx, -wy, -wz);
	modelView.translate(-position);
	dev->SetTransform(D3DTS_VIEW, (D3DMATRIX *) (const float *) transpose(modelView));

	dev->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);

	vec3 lightPos = vec3(300 * cosf(time), 150 * cosf(0.37f * time), 300 * sinf(time));

	drawDepth(lightPos);

	renderer->changeRenderTarget(rtColor, rtDepth);
    dev->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_RGBA(0, 0, 0, 0), 1.0f, 0);

	renderer->setShader(lighting);
	renderer->setTextures(base, cubeTarget);
	renderer->apply();

	renderer->changeShaderConstant4x4f("view_proj_matrix", projection * modelView);
	renderer->changeShaderConstant3f("lightPos", lightPos);
	renderer->changeShaderConstant3f("camPos", position);
	renderer->changeShaderConstant1f("distScale", DIST_SCALE);

	model->draw(dev);

	drawLight(lightPos);

	renderer->changeToDefaultRenderTarget();


	renderer->setShader(asciiShader);
	renderer->setTextures(rtColor, asciiTex);
	renderer->apply();

	renderer->changeShaderConstant1f("size", SIZE);
	renderer->changeShaderConstant1f("invSize", 1.0f / SIZE);

	drawQuad();

	return true;
}

App *app = new MainApp();
